#include "Renderer.h"
#include "Tools.h"
#include <algorithm>
#include <stdio.h>
#include "ShaderProgram.h"
#include "glm/gtc/type_ptr.hpp"
#include "glm/gtc/matrix_transform.hpp"

// RENDERER
Renderer::Renderer() : m_program("Geometry Rendering Program")
{
	m_vbo_vertices = 0;
	m_vbo_colors = 0;
	m_vao = 0;

	m_continous_time = 0.0;
}

Renderer::~Renderer()
{
	// delete common data
	glDeleteVertexArrays(1, &m_vao);
	glDeleteBuffers(1, &m_vbo_vertices);
	glDeleteBuffers(1, &m_vbo_colors);
}

bool Renderer::Init(int SCREEN_WIDTH, int SCREEN_HEIGHT)
{
	m_screen_width = SCREEN_WIDTH;
	m_screen_height = SCREEN_HEIGHT;

	// Initialize OpenGL functions

	//Set clear color
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	glDisable(GL_BLEND);
	
	// open the viewport
	glViewport(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT); //we set up our viewport

	bool techniques_initialization = InitRenderingTechniques();
	bool items_initialization = InitCommonItems();

	//If there was any errors
	if (Tools::CheckGLError() != GL_NO_ERROR)
	{
		printf("Exiting with error at Renderer::Init\n");
		return false;
	}

	//If everything initialized
	return techniques_initialization && items_initialization;
}

void Renderer::Update(float dt)
{
	m_continous_time += dt;
}

bool Renderer::InitRenderingTechniques()
{
	bool initialized = true;

	std::string vertex_shader_path = "../Assets/Shaders/basic_rendering.vert";
	std::string fragment_shader_path = "../Assets/Shaders/basic_rendering.frag";
	m_program.LoadVertexShaderFromFile(vertex_shader_path.c_str());
	m_program.LoadFragmentShaderFromFile(fragment_shader_path.c_str());
	m_program.CreateProgram();
	m_program.LoadUniform("uniform_color");
	
	return initialized;
}

bool Renderer::ReloadShaders()
{
	bool reloaded = true;
	// rendering techniques
	m_program.ReloadProgram();
	
	return reloaded;
}

bool Renderer::ResizeBuffers(int width, int height)
{
	m_screen_width = width;
	m_screen_height = height;

	return true;
}

void Renderer::Render()
{
	// Draw the geometry
	RenderGeometry();

	GLenum error = Tools::CheckGLError();
	if (error != GL_NO_ERROR)
	{
		printf("Reanderer:Draw GL Error\n");
		system("pause");
	}
}

#define TEST_CASE 1

bool Renderer::InitCommonItems()
{
	glGenVertexArrays(1, &m_vao);
	glBindVertexArray(m_vao);

	glGenBuffers(1, &m_vbo_vertices);
	glBindBuffer(GL_ARRAY_BUFFER, m_vbo_vertices);

#if TEST_CASE == 1
	// TEST: 1
	float vertices[] = {
		 0.0,   0.5, 0.0,
		-0.5,  -0.5, 0.0,
		 0.5,  -0.5, 0.0
	};
	
	m_number_of_vertices = 3;
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);

	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);

#elif TEST_CASE == 2

	// TEST: Using GLM
	glm::vec3 topVertex(0, 0.5, 0.0);
	glm::vec3 leftVertex(-0.5, -0.5, 0.0);
	glm::vec3 rightVertex(0.5, -0.5, 0.0);
	std::vector<glm::vec3> vertices = { topVertex, leftVertex, rightVertex };
	m_number_of_vertices = vertices.size();
	m_number_of_vertices = 3;
	glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(glm::vec3), &vertices[0][0], GL_STATIC_DRAW);

	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);
#elif TEST_CASE == 3

	// TEST: Triangle Strip
	// triangle strip
	std::vector<glm::vec3> vertices = {
		glm::vec3(-0.5, 0.5, 0.0),
		glm::vec3(-0.5, -0.5, 0.0),
		glm::vec3(0.0, 0.5, 0.0),
		glm::vec3(0.0, -0.5, 0.0),
		glm::vec3(0.5, 0.5, 0.0),
		glm::vec3(0.5, -0.5, 0.0)
	};
	m_number_of_vertices = vertices.size();
	glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(glm::vec3), &vertices[0][0], GL_STATIC_DRAW);


	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);

#elif TEST_CASE == 4

	// TEST: Color attribute
	glm::vec3 topVertex(0, 0.5, 0.0);
	glm::vec3 leftVertex(-0.5, -0.5, 0.0);
	glm::vec3 rightVertex(0.5, -0.5, 0.0);
	std::vector<glm::vec3> vertices = { topVertex, leftVertex, rightVertex };
	m_number_of_vertices = vertices.size();
	glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(glm::vec3), &vertices[0][0], GL_STATIC_DRAW);

	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);

	std::vector<glm::vec3> colors = { glm::vec3(1,0,0), glm::vec3(0,1,0), glm::vec3(0,0,1) };
	glGenBuffers(1, &m_vbo_colors);
	glBindBuffer(GL_ARRAY_BUFFER, m_vbo_colors);
	glBufferData(GL_ARRAY_BUFFER, colors.size() * sizeof(glm::vec3), &colors[0][0], GL_STATIC_DRAW);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, 0);
	
#endif

	glBindVertexArray(0);
	return true;
}

#define DRAW_CASE 2

void Renderer::RenderGeometry()
{
	// render to screen                    **********************************************
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glViewport(0, 0, m_screen_width, m_screen_height);
	
	glClear(GL_COLOR_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);

	m_program.Bind();

	glm::vec3 color = glm::vec3(1, 0, 0);
	glUniform3f(m_program["uniform_color"], color.r, color.g, color.b);
	
	glBindVertexArray(m_vao);

#if DRAW_CASE == 1
	glDrawArrays(GL_POINTS, 0, m_number_of_vertices);
	glPointSize(10);
#elif DRAW_CASE == 2
	glDrawArrays(GL_TRIANGLES, 0, m_number_of_vertices);
#elif DRAW_CASE == 3
	glDrawArrays(GL_TRIANGLE_STRIP, 0, m_number_of_vertices);
#elif DRAW_CASE == 4
	glDrawArrays(GL_LINES, 0, m_number_of_vertices);
#elif DRAW_CASE == 5
	glDrawArrays(GL_LINE_STRIP, 0, m_number_of_vertices);
#elif DRAW_CASE == 6
	glDrawArrays(GL_LINE_LOOP, 0, m_number_of_vertices);
#endif

	glBindVertexArray(0);

	m_program.Unbind();
}